/*******************************************************************
 * Bertha.h
 * target: PushpinV3
 *
 * Josh Lifton
 * lifton@media.mit.edu
 * MIT Media Lab
 * Copyright 2002 all rights reserved.
 *
 * Last modified 07MAR2002 by lifton.
 *
 * This file defines the memory layout for the Bertha operating 
 * system as implemented on the Cygnal C8051F016 microprocessor.
 *******************************************************************/

/*
 * Enumerations used throughout the operating system.
 */

enum packetTypes {
	DIRECTTOFLASH=0, OSHEADER, OFF, SETRANDOM, 
	GRADIENT, ELECTLOWEST, REQUESTVOTE, VOTE, LOCALRESULT, 
	GLOBALRESULT, COLOR
};

enum serialStates {
	DISABLED, IDLE, SCANNING, SENDINGTOADDRESS, RECEIVINGHEADER, 
	RECEIVINGCONTENT, IGNORINGCONTENT, SENDINGHEADER, SENDINGCONTENT, 
	PACKETRECEIVED, PACKETSENT, STASIS
};

enum commModes {
	DEFAULTCOMMMODE=1, 
	RS232IRCONVERTER_RS232=1, 
	RS232IRCONVERTER_IR, 
	FOURWAYIR		
};

enum converterSelectOptions {
	RS232SELECT=0, IRSELECT=1
};

enum QueuePacketReturnCodes {
	PACKETNOTQUEUED,
	PACKETQUEUED,
	PACKETTRANSMITTING
};

enum LEDcolors {
	RED1, RED2, AMBER1, YELLOW1, GREEN1, NOCOLOR
};

/*
 * The start address of the portion of flash memory used as swap 
 * space for temporarily holding a new version of the OS.
 */
#define swapSpace ((unsigned char code *) 0x3C00) // swapSpace and OS space should be equal = 0x7800/2
#define swapSpaceSize ((unsigned int) 0x3C00)

/*
 * Constants and definitions specific to the microprocessor 
 * hardware.  Either 8051-specific or C8051F016-specific.
 */
#define NUMBEROFADCCHANNELS ((unsigned char) 9) /* Temp sensor + 8 pinned out.       */
#define COMMADCCHANNEL ((unsigned char) 0)      /* ADC channel on comm module.       */
#define IRADCTHRESHOLD ((unsigned int) 0x0200) /* ADC value for which IR detected.  */
#define XdataSize ((unsigned int) 2048)        /* 2Kbytes external memory (RAM).    */
#define FlashSize ((unsigned int) 32896)	      /* Flash memory size.             */
#define FlashPageSize ((unsigned int) 512)	  /* Minimum flash memory erase size.  */
#define MiniPageAddress ((unsigned int) 0x8000) /* Extra flash at top of memory. */
#define MiniPageSize ((unsigned int) 128)	/* Size of extra flash at top of memory.*/
#define ErasedPFragChar ((unsigned char)0xff)  /* Erased flash overwrite character. */
typedef unsigned char DataAddress;        /* 256 bytes.            */
typedef unsigned char IdataAddress;       /* 256 bytes.            */
typedef unsigned int XdataAddress;        /* 2048 bytes.           */
typedef unsigned int CodeAddress;         /* 32 Kbytes.            */
typedef unsigned char ReservedMemoryArea; /* Smallest memory unit. */
typedef union AddressableLong {           /* Utility data type.    */
   unsigned long Long;   /* Access as a single unsigned long.      */
   unsigned int Int[2];  /* Access as individual unsigned integers.*/
   unsigned char Char[4];/* Access as individual unsigned bytes.   */
} AddressableLong;
typedef union AddressableInt {            /* Utility data type.    */
   unsigned int Int;     /* Access as individual unsigned integers.*/
   unsigned char Char[2];/* Access as individual unsigned bytes.   */
} AddressableInt;

/*
 * Timing constants.  Determined in large part by the bandwidth 
 * available for communication, 92160 baud, which translates to 
 * approximately 4600 bytes/second for the 4-way IR communication
 * module.  The largest possible PFrag and associated PFragState 
 * is approximately 2500 bytes.  
 */
#define swapSpaceEraseTime ((unsigned int) 354)  /* milliseconds needed to erase the swapSpace (0x3C00)*/
#define ONESECOND ((unsigned long) 1843318)
#define ONEMILLISECOND ((unsigned long) 1844)
#define DELAYTIME ((unsigned int) 0xFFFF)     /* This is about 32.6 milliseconds. */
#define minIDVetoWaitTime ((unsigned long) 3 * (unsigned long) timeScale * (unsigned long) DELAYTIME)
#define minNeighborhoodBuildTime ((unsigned long) 0xFFFF)
#define DEFAULTTIMESCALE ((unsigned int) 30)
#define NEIGHBORTIMEOUT ((unsigned long) (5 * timeScale * DELAYTIME))

#define maxPFragSize ((unsigned int) 2048)		/* A pfrag's code is stored in one 2Kbyte sector of flash. */

typedef unsigned char PushpinLocalID;

typedef struct BerthaPacket {
	PushpinLocalID toAddress;	// Intended recipient's neighborhood ID.
	PushpinLocalID fromAddress;	// Sender's neighborhood ID.
	unsigned char packetType;	// Classification of this packet.
	XdataAddress contentSize;	// The number of bytes of content contained in this packet.
	unsigned char CRC;			// Cyclic redundancy check sum to allow for error detection.
	unsigned char xdata * contentPtr;	// A pointer to the content of this packet.
} Packet;
#define PACKETHEADERSIZE ((unsigned int) 6)
#define GLOBALADDRESS ((unsigned char) 0)
#define IDEADDRESS ((unsigned char) 0xFF)

typedef struct OSHeaderDefinition {
	unsigned int size;
	unsigned int version;
	unsigned char crc;
} OSHeader;

typedef struct OSStateDefinition {
	unsigned int randomSeed;
	unsigned int size;
	unsigned int version;
	unsigned char isOld;
	unsigned char crc;
} OSState;

#define maxNumberOfCandidates ((unsigned char) 5)

typedef union MessageContentDefinition {
	OSHeader osHeader;
	OSState osState;
	unsigned int candidate;
	unsigned int candidateList[maxNumberOfCandidates];
	unsigned int electionResult;
	unsigned char color;
	unsigned int randomSeed;
} MessageContent;

typedef struct MessageDefinition {
//	unsigned char uid;					// An ID assigned by the Pushpin of origin.
	unsigned char hopCount;				// The number of Pushpins this message has visited since the Pushpin of origin.
	unsigned char maxHopCount;			// The furthest from the Pushpin of origin this message should travel.
//	unsigned long timeStamp;			// Local time at the transmitting Pushpin.
//	unsigned char timesToRebroadcast;
//	unsigned char lowerTimerBound;
//	unsigned char upperTimerBound;
	MessageContent content;
} Message;

typedef struct QueuedMessageDefinition {
	unsigned char to;
	unsigned char type;
	unsigned int size;
	unsigned char copiesToSend;
	unsigned long timeBetweenCopies;
	unsigned long timeNextToSend;
	Message message;
} QueuedMessage;

#define MessageQueueSize 10		// Should be less than 0xFF.
#define IncomingBufferSize (sizeof(Message))
#define OutgoingBufferSize (sizeof(Message))

/* Internal RAM Layout (256 bytes)
 *
 *	----------------------------------------------------
 *  |  Stack                  Addresses 192 - 255      |
 *  |   (automatically allocated by the compiler)      |
 *	|--------------------------------------------------|
 *	|  Active PFrag State     Addresses 190 - 191      |
 *  |    Pointer                                       |
 *	|--------------------------------------------------|
 *  |  Active PFrag Local     Addresses 32 - 189       |
 *  |    Variables                                     |
 *	|--------------------------------------------------|
 *	|  Registers              Addresses 0 - 31         |
 *	----------------------------------------------------
 */
#define ActivePFragStatePointerAddress ((unsigned int) 190) /* A two-byte external memory pointer. */


/* External RAM Layout (2Kbytes) 
 *	----------------------------------------------------
 *  |  Reentrant Stack        Addresses 1998 - 2047    |
 *  |--------------------------------------------------|
 *	|  Neighborhood Watch     Addresses 1152 - 1997    |
 *  |--------------------------------------------------|
 *  |  Bulletin Board System  Addresses 576 - 1151     |
 *	|--------------------------------------------------|
 *	|  PFrag State Table      Addresses 128 - 575      |
 *	|--------------------------------------------------|
 *  |  OS Local Variables     Addresses 0 - 127        |
 *	----------------------------------------------------
 */
//#define maxPFragStateTableSize ((unsigned int) 448)	/* A pfrag's state is stored within the 2Kbytes of external RAM. */
//#define maxPFragStateSize ((unsigned int) 256) 		/* Limit size specification to one byte. */
//#define maxBBSSize ((unsigned int) 576)				/* Amount of RAM available for pfrags to exchange information. */
//#define maxBBSPostSize maxBBSSize	/* One large message could fill the BBS. */
//#define maxNWSize ((unsigned int) 846)     			/* All neighbor information must fit in here. */
//#define NoMorePFrags ((unsigned char) 0xFF)			/* Indicates end of the state table. */
//#define NullPFragUID ((unsigned int) 0xFFFF)			/* A PFrag with this UID is not valid. */
/* No linked list can be located at '0'!*/
//#define PFragStateTableAddress ((unsigned int) 0x80)	/* Location of the table containing all PFrag states. */
//#define BBSAddress ((unsigned int) 0x240)			/* PFragStateTableAddress + PFragStateTableSize */
//#define NWAddress ((unsigned int) 0x480)				/* BBSAddress + maxBBSSize */

/* Program Memory Layout (32Kbytes Flash)
 *
 *	----------------------------------------------------
 *  |  Random Seed            Addresses 32768 - 32889  |
 *  |--------------------------------------------------|
 *  |  Reserved               Addresses 32256 - 32767  |
 *  |    (Security Lock)                               |
 *  |--------------------------------------------------|
 *  |  Lookup Tables          Addresses 31232 - 32255  |
 *  |    (codecs, CRC, etc.)                           |
 *  |--------------------------------------------------|
 *  |  OSOVERWRITE functions  Addresses 30720 - 31231  |
 *  |--------------------------------------------------|
 *  |  Code Fragment #9       Addresses 28672 - 30719  |
 *  |--------------------------------------------------|
 *  |  Code Fragment #8       Addresses 26624 - 28671  |
 *  |--------------------------------------------------|
 *  |  Code Fragment #7       Addresses 24576 - 26623  |
 *  |--------------------------------------------------|
 *  |  Code Fragment #6       Addresses 22528 - 24575  |
 *  |--------------------------------------------------|
 *  |  Code Fragment #5       Addresses 20480 - 22527  |
 *  |--------------------------------------------------|
 *  |  Code Fragment #4       Addresses 18432 - 20479  |
 *  |--------------------------------------------------|
 *  |  Code Fragment #3       Addresses 16384 - 18431  |
 *  |--------------------------------------------------|
 *  |  Code Fragment #2       Addresses 14336 - 16383  |
 *  |--------------------------------------------------|
 *  |  Code Fragment #1       Addresses 12288 - 14335  |
 *  |--------------------------------------------------|
 *	|  Operating System       Addresses 0000 - 12287   |
 *	----------------------------------------------------
 */

// Bit twiddling macros for Neighbor.bitFlags.
#define DEAD (0x01)             // This neighbor should be removed if set.

#define isNeighborDead(n) ((((Neighbor xdata *) n) -> bitFlags) & DEAD)
#define setNeighborDead(n) ((((Neighbor xdata *) n) -> bitFlags) |= DEAD)
#define clearNeighborDead(n) ((((Neighbor xdata *) n) -> bitFlags) &= ~DEAD)

// Bit twiddling macros for PFragState.bitFlags.
#define VALID (0x01)            // This is a valid PFrag.
#define PFRAG (0x02)            // This contains a completely received, but unvalidated PFrag.
#define INTRANSIT (0x04)        // A new PFrag is coming in here over the comm channel.
#define INSTALLED (0x08)        // This PFrag has been installed.
#define READYTOREMOVE (0x10)    // This PFrag should be deleted.
#define STATEINITIALIZED (0x20) // This PFrag came with a valid state.

#define isPFragValid(id) (PFragStatus[id] & VALID)
#define setPFragValid(id) (PFragStatus[id] |= VALID)
#define clearPFragValid(id) (PFragStatus[id] &= ~VALID)

#define isPFrag(id) (PFragStatus[id] & PFRAG)
#define setPFrag(id) (PFragStatus[id] |= PFRAG)
#define clearPFrag(id) (PFragStatus[id] &= ~PFRAG)

#define isPFragInTransit(id) (PFragStatus[id] & INTRANSIT)
#define setPFragInTransit(id) (PFragStatus[id] |= INTRANSIT)
#define clearPFragInTransit(id) (PFragStatus[id] &= ~INTRANSIT)

#define isPFragInstalled(id) (PFragStatus[id] & INSTALLED)
#define setPFragInstalled(id) (PFragStatus[id] |= INSTALLED)
#define clearPFragInstalled(id) (PFragStatus[id] &= ~INSTALLED)

#define isPFragRemoveReady(id) (PFragStatus[id] & READYTOREMOVE)
#define setPFragRemoveReady(id) (PFragStatus[id] |= READYTOREMOVE)
#define clearPFragRemoveReady(id) (PFragStatus[id] &= ~READYTOREMOVE)

#define isPFragStateInitialized(id) (PFragStatus[id] & STATEINITIALIZED)
#define setPFragStateInitialized(id) (PFragStatus[id] |= STATEINITIALIZED)
#define clearPFragStateInitialized(id) (PFragStatus[id] &= ~STATEINITIALIZED)


// ADC functions.
#define getADCChannel() (AMX0SL)
#define setADCChannel(channel) (AMX0SL = channel)
#define clearADCWindowInterruptFlag() (ADWINT = 0)
#define enableADCWindowInterrupt() (EIE1 |= 0x04)
#define disableADCWindowInterrupt() (EIE1 &= ~(0x04))
sfr16 ADCWindowGreaterThan = 0xC4;
sfr16 ADCWindowLessThan = 0xC6;

// SFR extra bits and redefinitions.
sfr16 timer3reload = 0x92;
sfr16 timer3 = 0x94;
#define isBuildingNeighborhood() (PCON & 0x04)
#define setBuildingNeighborhood() (PCON |= 0x04)
#define clearBuildingNeighborhood() (PCON &= ~(0x04))
#define isPacketInQueue() (PCON & 0x08)
#define setPacketInQueue() (PCON |= 0x08)
#define clearPacketInQueue() (PCON &= ~(0x08))
#define isChannelClear() (PCON & 0x10)
#define setChannelClear() (PCON |= 0x10)
#define clearChannelClear() (PCON &= ~(0x10))
//#define isHighAlarmArmed() (PCON & 0x20)
//#define setHighAlarmArmed() (PCON |= 0x20)
//#define clearHighAlarmArmed() (PCON &= ~(0x20))
//#define isNWBusy() (PCON & 0x40)
//#define setNWBusy() (PCON |= 0x40)
//#define clearNWBusy() (PCON &= ~(0x40))

// Programmable Counter Array functions.
#define enableAllPCAInterrupts() EIE1 |= 0x08; PCA0MD |= 0x01
#define disableAllPCAInterrupts() EIE1 &= ~(0x08); PCA0MD &= ~(0x01)
#define clearPCAInterruptFlag() CF = 0;
#define clearPCA4InterruptFlag() CCF4 = 0
#define enablePCA4Interrupts() PCA0CPM4 |= 0x01
#define disablePCA4Interrupts() PCA0CPM4 &= ~(0x01)
#define enablePCA4MatchInterrupt() PCA0CPM4 |= 0x08
#define disablePCA4MatchInterrupt() PCA0CPM4 &= ~(0x08)

// Flash memory functions.
#define enableFlashErase() PSCTL = 0x03
#define disableFlashErase() PSCTL = 0x00
#define enableFlashWrite() PSCTL = 0x01
#define disableFlashWrite() PSCTL = 0x00

// Timer 0 functions.
#define enableTimer0() TR0 = 1
#define disableTimer0() TR0 = 0
#define enableTimer0Interrupt() ET0 = 1
#define disableTimer0Interrupt() ET0 = 0
#define clearTimer0InterruptFlag() TF0 = 0
#define timer0FromSystemClock() CKCON |= 0x08
#define timer0FromSystemClockDiv12() CKCON &= ~(0x08)

// Timer 1 functions.
#define enableTimer1() TR1 = 1
#define disableTimer1() TR1 = 0
#define enableTimer1Interrupt() ET1 = 1
#define disableTimer1Interrupt() ET1 = 0
#define clearTimer1InterruptFlag() TF1 = 0

// Timer 2 functions.
#define enableTimer2() TR2 = 1
#define disableTimer2() TR2 = 0
#define enableTimer2Interrupt() ET2 = 1
#define disableTimer2Interrupt() ET2 = 0
#define clearTimer2InterruptFlag() TF2 = 0

// Timer 3 functions.
#define enableTimer3() TMR3CN |= 0x04
#define disableTimer3() TMR3CN &= ~(0x04)
#define enableTimer3Interrupt() EIE2 |= 0x01
#define disableTimer3Interrupt() EIE2 &= ~(0x01)
#define clearTimer3InterruptFlag() TMR3CN &= ~(0x80)
#define timer3FromSystemClock() TMR3CN |= 0x02
#define timer3FromSystemClockDiv12() TMR3CN &= ~(0x02)

// General interrupt functions.
#define enableGlobalInterrupts() EA = 1
#define disableGlobalInterrupts() EA = 0

// Serial UART functions.
#define enableSerialInterrupts() ES = 1
#define disableSerialInterrupts() ES = 0
#define enableSerialReceive() REN = 1
#define disableSerialReceive() REN = 0
#define setTransmitInterruptFlag() TI = 1
#define clearTransmitInterruptFlag() TI = 0
#define clearReceiveInterruptFlag() RI = 0
#define listenToAddressBytes() SM2 = 1
#define listenToAllBytes() SM2 = 0
#define setUARTMode0() SM0 = 0; SM1 = 0
#define setUARTMode1() SM0 = 0; SM1 = 1
#define setUARTMode2() SM0 = 1; SM1 = 0
#define setUARTMode3() SM0 = 1; SM1 = 1
#define sendAddressBytes() TB8 = 1
#define sendDataBytes() TB8 = 0

// meme : may not need these.
// Serial Peripheral Interface functions.
#define disableSPI() SPIEN = 0
#define enableSPIInterrupt() EIE1 |= 0x01
#define disableSPIInterrupt() EIE1 &= ~(0x01)
#define setSPIInterruptFlag() SPIF = 1
#define clearSPIInterruptFlag() SPIF = 0

// External interrupt functions.
#define enableExternal4Interrupt() (EIE2 |= 0x04)
#define enableExternal5Interrupt() (EIE2 |= 0x08)
#define enableExternal6Interrupt() (EIE2 |= 0x10)
#define enableExternal7Interrupt() (EIE2 |= 0x20)
#define disableExternal4Interrupt() (EIE2 &= ~(0x04))
#define disableExternal5Interrupt() (EIE2 &= ~(0x08))
#define disableExternal6Interrupt() (EIE2 &= ~(0x10))
#define disableExternal7Interrupt() (EIE2 &= ~(0x20))
#define clearExternal4InterruptFlag() (PRT1IF &= ~(0x10))
#define clearExternal5InterruptFlag() (PRT1IF &= ~(0x20))
#define clearExternal6InterruptFlag() (PRT1IF &= ~(0x40))
#define clearExternal7InterruptFlag() (PRT1IF &= ~(0x80))

// Software-forced reset.
#define reboot() (RSTSRC |= 0x02)