/*******************************************************************************
 * BerthaTiming.c
 * target: PushpinV3
 *
 * Josh Lifton
 * lifton@media.mit.edu
 * MIT Media Lab
 * Copyright 2002 all rights reserved.
 *
 * Last modified 08APR2002 by lifton.
 *
 * Implementation of Bertha's timing functions, including the real-time clock.
 * Assumes the Pushpin is running off a 22.1184MHz crystal.  A 32-bit real-time
 * clock is formed by conjoining timer 0 running in 16-bit clock mode with a 
 * 16-bit unsigned integer incremented every time timer 0 overflows.  Timer 0 
 * is configured increment once for every 12 system clocks.  Given this 
 * arrangement, the 32-bit real-time clock should overflow once every 38.84 
 * minutes.  The observed overflow rate, however, is approximately twice this 
 * amount, perhaps due to how system cycles are counted.
 *
 * Software alarms can be set such that an interrupt is generated after a 
 * specified amount of time, up to the limit of the real-time clock.  Timer 3  
 * triggers the actual alarm.  Overhead between calling the setAlarm function 
 * and actually setting the alarm accounts for approximately 32 microseconds.  
 * This overhead time is taken into account when setting the alarm.
 *
 * This program demonstrates starting the real-time clock and setting an alarm.
 * Each time an alarm goes off, the LED on pin P1^3 toggles and another alarm is 
 * set.  The desired outcome is that the LED toggles at intervals determined by 
 * the alarm.  
 *******************************************************************************/

#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>

unsigned int data highTime;

void startRealTimeClock(void) {
    /*
     * Timer 0 is the lower 16 bits of the 32 bit real-time clock.
     */
    disableTimer0();
    enableTimer0Interrupt();
    clearTimer0InterruptFlag();
	timer0FromSystemClockDiv12();
    TMOD &= ~(0x0F);    // Clear timer 0 control registers.
    TMOD |= 0x01;       // Set timer 0 as 16-bit timer.
	highTime = 0;
    setTime(0);
    enableTimer0();
}

unsigned long getTime(void) {
    AddressableLong data time;
	disableTimer0();
	time.Int[0] = highTime;
    time.Char[2] = TH0;
    time.Char[3] = TL0;
	enableTimer0();
    return time.Long;
}

void setTime(unsigned long time) {
    TL0 = *(((unsigned char *) &time) + 3);
    TH0 = *(((unsigned char *) &time) + 2);
	highTime = *((unsigned int *) &time);
}

void timer0Interrupt(void) interrupt 1 {
    clearTimer0InterruptFlag();
	highTime++;
}