/*******************************************************************************
 * Messaging.c
 * target: PushpinV3
 *
 * Josh Lifton
 * lifton@media.mit.edu
 * MIT Media Lab
 * Copyright 2002 all rights reserved.
 *
 * Last modified 23NOV2002 by lifton.
 *
 * Functions for higher level messaging.  
 *******************************************************************************/

#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <Bertha.h>
#include <BerthaFunctionPrototypes.h>
#include <BerthaGlobals.h>
#include <stdlib.h>
#include <string.h>

/*
 * Messaging variables.
 */

QueuedMessage xdata MessageQueue[MessageQueueSize];
unsigned char currentMessage = 0xFF;

/*
 * Global variables.
 */
OSHeader xdata incomingOSHeader = {0,0,0};
unsigned char xdata inLight = 0;

/*
 * Handles incoming messages delivered from neighboring Pushpins.
 * Assumes the message itself is stored in IncomingBuffer and 
 * the header information is stored in receivePacket.  Also 
 * assumes the message and header are error-free.
 */
void processIncomingMessage(Message xdata * messagePtr) {
	/*
	 * Packets written directly to flash are dealt with differently 
	 * than all other packets.
	 */
	if (receivePacket.packetType == DIRECTTOFLASH) {
		/*
		 * Check if a new OS was just copied into the swap space.
		 */
		if (incomingOSHeader.size) {
			if (getFlashBlockCRC8(swapSpace,incomingOSHeader.size) == incomingOSHeader.crc) {
				/*
				 * Successfully received a new OS.
				 */
				((OSState xdata *) outgoingMessagePtr)->randomSeed = osState.randomSeed;
				((OSState xdata *) outgoingMessagePtr)->isOld = 0;
				((OSState xdata *) outgoingMessagePtr)->size = incomingOSHeader.size;
				((OSState xdata *) outgoingMessagePtr)->version = incomingOSHeader.version;
				((OSState xdata *) outgoingMessagePtr)->crc = incomingOSHeader.crc;
				saveOSState((OSState xdata *) outgoingMessagePtr);
				overwriteOS(0, swapSpace, incomingOSHeader.size);
			}
			else {
				/*
				 * The new OS was not received correctly.
				 */
				incomingOSHeader.size = 0;
				incomingOSHeader.version = 0;
				incomingOSHeader.crc = 0;
			}
		}
		return;
	}

	/*
	 * Increment this message's hop count by one.
	 */
	(messagePtr->hopCount)++;

	/*
	 * Indicate the message is valid and is being considered.
	 */
	flashLED(1,60000);

	/*
	 * Act on the message.
	 */
	switch (receivePacket.packetType) {
		case OSHEADER :
			/*
			 * Copy the OS header from the buffer for future reference.
			 */
			incomingOSHeader.size = (messagePtr->content).osHeader.size;
			incomingOSHeader.version = (messagePtr->content).osHeader.version;
			incomingOSHeader.crc = (messagePtr->content).osHeader.crc;
			/*
			 * Check if the incoming OS is a newer version.  Any version is newer than 0xFFFF.
			 * If so, check if there is enough space for the OS.  
			 * If so, clear it.  Otherwise, delete the OS header.
			 */
			if (incomingOSHeader.size < swapSpaceSize 
			    && (incomingOSHeader.version > osState.version || osState.version == 0xFFFF)) {
				clearFlashMemory(swapSpace, incomingOSHeader.size);
			}
			else {
				incomingOSHeader.size = 0;
				incomingOSHeader.version = 0;
				incomingOSHeader.crc = 0;
			}
			break;
		case OFF :
			clearLEDDisplay();
			break;
		case SETRANDOM :
			outgoingMessagePtr->content.osState.randomSeed = messagePtr->content.randomSeed;
			outgoingMessagePtr->content.osState.isOld = osState.isOld;
			outgoingMessagePtr->content.osState.size = osState.size;
			outgoingMessagePtr->content.osState.version = osState.version;
			outgoingMessagePtr->content.osState.crc = osState.crc;
			saveOSState(&(outgoingMessagePtr->content.osState));
			flashLED(1,40000);	//|
			delay(0xFFFF);		//| Confirm receipt of new random seed.
			flashLED(1,40000);	//|
			break;
		default :
			break;
	}
}
		
void checkMessageQueue(void) {
	unsigned char i;
	unsigned long time = getTime();
	/*
	 * Search for a queued message that should be sent if one isn't already being sent.
	 */
	if (!(currentMessage < MessageQueueSize)) {
		for (i=0; i<MessageQueueSize; i++) {
			if (MessageQueue[i].copiesToSend) {
				if (time > MessageQueue[i].timeNextToSend) {
				//	MessageQueue[i].message.timeStamp = time;
					if (queuePacketForTransmission(MessageQueue[i].to, 
													MessageQueue[i].type, 
    	             	                      		MessageQueue[i].size, 
													(unsigned char xdata *) &(MessageQueue[i].message))) {
						currentMessage = i;
					}
				}
			}
		}
	}
}

unsigned char findMessage(unsigned char type) {
	unsigned char i;
	for (i=0; i<MessageQueueSize; i++) {
		if (MessageQueue[i].type == type) {
			return i;
		}
	}
	return 0xFF;
}
		

void addToMessageQueue(unsigned char to, 
                       unsigned char type, 
					   unsigned int size, 
					   unsigned char copiesToSend, 
					   unsigned long timeBetweenCopies,  
					   Message xdata * messagePtr) {
	unsigned char i;
	unsigned int j;
	/*
	 * Search for an empty spot to put the message.
	 */
	for (i=0; i<MessageQueueSize; i++) {
		if (MessageQueue[i].copiesToSend == 0) {
			/*
			 * Found a spot to put the message, now copy it over.
			 */
			for (j=0; j<sizeof(Message); j++) {
				*(((unsigned char xdata *) &(MessageQueue[i].message)) + j ) = *(((unsigned char xdata *) messagePtr) + j);
			}
			/*
			 * Determine this message's metadata used for retransmission.
			 */
			MessageQueue[i].to = to;
			MessageQueue[i].type = type;
			MessageQueue[i].size = size;
			MessageQueue[i].copiesToSend = copiesToSend;
			MessageQueue[i].timeBetweenCopies = timeBetweenCopies;
			MessageQueue[i].timeNextToSend = getTime() + timeBetweenCopies;
			return;
		}
	}
}

void removeFromMessageQueue(unsigned char messageIndex) {
	unsigned char i;
	if (messageIndex < MessageQueueSize) {
		for (i=0; i<sizeof(QueuedMessage); i++) {
			*(((unsigned char xdata *) &MessageQueue[messageIndex]) + i) = 0;
		}
	}
}

void initializeMessaging(void) {
	unsigned int i;
	for (currentMessage=0; currentMessage<MessageQueueSize; currentMessage++) {
		for (i=0; i<sizeof(QueuedMessage); i++) {
			*((unsigned char xdata *) (((unsigned char xdata *) &(MessageQueue[currentMessage])) + i)) = 0;
		}
	}
	currentMessage = 0xFF;
}

void initializeMessage(Message xdata * m) {
	unsigned int i;
	for (i=0; i<(sizeof(Message)); i++) {
		*((unsigned char xdata *) (((unsigned int) m) + i)) = 0;
	}
}

/*
 * This function is only called after a packet has been sent and the 
 * communication system is ready to send another packet.  This is
 * the mechanism that allows for multi-packet messages to be sent.
 */
void nextMessage(void) {
	if (transmitPacket.packetType == OSHEADER) {
		delay(0xFFFF);
		delay(0xFFFF);
		delay(0xFFFF);
		delay(0xFFFF);
		delay(0xFFFF);	// Time needed to erase the swapSpace (0x3C00 bytes).
		delay(0xFFFF);
		delay(0xFFFF);
		delay(0xFFFF);
		delay(0xFFFF);
		delay(0xFFFF);
		delay(0xFFFF);
		setCommMode(RS232IRCONVERTER_IR);
		queuePacketForTransmission(transmitPacket.toAddress, 
				DIRECTTOFLASH,
				osState.size,
				(unsigned char xdata *) 0);
	}
	if (currentMessage < MessageQueueSize) {
		/*
		 * The message just sent exists in the MessageQueue
		 * and should be dealt with accordingly.
		 */
		if (--(MessageQueue[currentMessage].copiesToSend)) {
			MessageQueue[currentMessage].timeNextToSend = getTime() + MessageQueue[currentMessage].timeBetweenCopies;
		}
		else {
			removeFromMessageQueue(currentMessage);
		}
		/*
		 * No messages from the MessageQueue are currently being sent.
		 */
		currentMessage = 0xFF;
	}
}

/* 
 * For now, assume all communication is over IR.
 */
void propagateOS(void) {
	if (osState.size <= swapSpaceSize) {
		if (!(osState.isOld)) {
			/*
			 * This OS is no longer new
			 */
			((OSState xdata *) outgoingMessagePtr)->randomSeed = osState.randomSeed;
			((OSState xdata *) outgoingMessagePtr)->isOld = 1;
			((OSState xdata *) outgoingMessagePtr)->size = osState.size;
			((OSState xdata *) outgoingMessagePtr)->version = osState.version;
			((OSState xdata *) outgoingMessagePtr)->crc = osState.crc;
			saveOSState((OSState xdata *) outgoingMessagePtr);
		}
		(outgoingMessagePtr->content).osHeader.size = osState.size;
		(outgoingMessagePtr->content).osHeader.version = osState.version;
		(outgoingMessagePtr->content).osHeader.crc = osState.crc;
		flashLED(100,0x25000);
		setCommMode(RS232IRCONVERTER_IR);
		queuePacketForTransmission(GLOBALADDRESS, 
				OSHEADER,
				sizeof(Message) - sizeof(MessageContent) + sizeof(OSHeader),
				(unsigned char xdata *) outgoingMessagePtr);
	}
}
